/*********************************************************************************
 *  xksclient version 1.10, Informatica s.r.o, 1998
 *
 *  Published information of Informatica s.r.o
 *
 *  The INFORMATICA END USER LICENSE AGREEMENT is the document governing
 *  your use of Informatica software.
 *
 *  Please note that the LIMITED WARRANTY defined in Part II thereof
 *  applies only to the compiled form of this program as shipped by
 *  Informatica, not to this source code module.
 *
 *  DISCLAIMER OF WARRANTY
 *
 *  THIS SOURCE CODE MODULE IS PROVIDED ON AN "AS IS" BASIS, IN THE HOPE THAT
 *  IT WILL BE USEFUL FOR REFERENCE OR AS A BASIS FOR FURTHER DEVELOPMENT,
 *  OR INTEGRATION INTO OTHER SYSTEMS, PROGRAMS OR APPLICATIONS,
 *  WITHOUT WARRANTY OF ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES
 *  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
 *  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS BORNE BY YOU.
 *  SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU AND NOT INFORMATICA ASSUME THE ENTIRE
 *  COST OF ANY SERVICE AND REPAIR. IN ADDITION, THE SECURITY MECHANISMS
 *  IMPLEMENTED BY INFORMATICA SOFTWARE HAVE INHERENT LIMITATIONS, AND YOU MUST
 *  DETERMINE THAT THE SOFTWARE SUFFICIENTLY MEETS YOUR REQUIREMENTS.
 *  SOME JURISDICTIONS DO NOT ALLOW EXCLUSIONS OF AN IMPLIED WARRANTY,
 *  SO THIS DISCLAIMER MAY NOT APPLY TO YOU AND YOU MAY HAVE OTHER
 *  LEGAL RIGHTS THAT VARY BY JURISDICTION.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#ifndef USE_XAW

#include <Xm/ToggleBG.h>
#include <Xm/RowColumn.h>

#else

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Toggle.h>

#endif


/*********************************************************************************
 *  Program identifier (the only global data)
 */
char *identifier = "xksclient version 1.10, Informatica s.r.o, 1998";


/*********************************************************************************
 *  Configuration and other useful constants
 */
#define APPLICATION_CLASS	"Xksclient"		/* application class name for X Toolkit */
#define RADIO_BOX		"radio_box"		/* name of the RadioBox widget */
#define DEFAULT_PATH		":/bin:/usr/bin"	/* if PATH doesn't exist */
#define DISPLAY_PROPERTY	"XKS_DISPLAY"		/* name of property containing X client conn info */
#define KEYBOARD_PROPERTY	"XKS_KEYBOARD"		/* name of property containing keyboard ctl conn info */
#define HOST_LEN		1024			/* maximum length of a hostname */
#define LINK_LEN		1024			/* maximum length of a symlink */
#define DISP_LEN		256			/* maximum length of DISPLAY-like variables */


/*********************************************************************************
 *  Preprocessor macros
 *
 *  OFFSET calculates the offset of a given field in the structure
 *  SYS_ERROR prints a system-specific error message to stderr (errno is evaluated)
 *  APP_ERROR prints an application-specific error message to stderr
 *
 *  The other macros are used to produce a debug version using the -DDEBUGGING switch
 */
#define OFFSET( type_p, field ) ((int) ((char *) &((( type_p ) NULL ) -> field )))
#define SYS_ERROR( fmt ) { fprintf( stderr, "%s: %d: ", __FILE__, __LINE__ ); fprintf( stderr, fmt ); perror( " " ); }
#define SYS_ERROR1( fmt, arg1 ) { fprintf( stderr, "%s: %d: ", __FILE__, __LINE__ ); fprintf( stderr, fmt, arg1 ); perror( " " ); }
#define APP_ERROR( fmt ) { fprintf( stderr, "%s: %d: ", __FILE__, __LINE__ ); fprintf( stderr, fmt ); fprintf( stderr, "\n" ); }
#define APP_ERROR1( fmt, arg1 ) { fprintf( stderr, "%s: %d: ", __FILE__, __LINE__ ); fprintf( stderr, fmt, arg1 ); fprintf( stderr, "\n" ); }

#ifdef DEBUGGING

#define DEBUGNL() { fprintf( stdout, "\n" ); }
#define DEBUGN( fmt ) { fprintf( stdout, "DEBUG: " ); fprintf( stdout, fmt ); }
#define DEBUG( fmt ) { fprintf( stdout, "DEBUG: " ); fprintf( stdout, fmt ); fprintf( stdout, "\n" ); }
#define DEBUG1( fmt, arg1 ) { fprintf( stdout, "DEBUG: " ); fprintf( stdout, fmt, arg1 ); fprintf( stdout, "\n" ); }
#define DEBUG2( fmt, arg1, arg2 ) { fprintf( stdout, "DEBUG: " ); fprintf( stdout, fmt, arg1, arg2 ); fprintf( stdout, "\n" ); }

#else

#define DEBUGNL()
#define DEBUGN( fmt )
#define DEBUG( fmt )
#define DEBUG1( fmt, arg1 )
#define DEBUG2( fmt, arg1, arg2 )

#endif


/*********************************************************************************
 **	DATATYPES USED IN THIS PROGRAM						**
 ********************************************************************************/

/*
 *  Keyboard types are kept in a linked list with elements of type KType
 */
typedef struct {
	char *next;			/* pointer to next structure in the linked list */
	char *config;			/* pointer to the Config structure */
	char *name;			/* name used to identify this keyboard type */
	char *nice_name;		/* name with underscores replaced by spaces */

	Widget toggle;			/* ToggleButton for this keyboard type */
} KType;


/*
 *  Config holds all global information in this program
 */
typedef struct {
	KType *first_ktype;		/* pointer to the list of keyboard types */
	KType *active_ktype;		/* pointer to the active keyboard type */

	int ksock;			/* connected socket to Xks */
} Config;

/*
 *  Function prototypes generated externally
 *  If you cannot generate function prototypes, just delete this line
 */
#include "xksclient.cproto"


/*********************************************************************************
 **		UTILITIES FOR KEYBOARD CONTROL PROTOCOL HANDLING		**
 ********************************************************************************/

/*********************************************************************************
 *  FIND_WORD()
 *
 *  FIND_WORD() searches for word by skipping over separator characters.
 *  The carriage return is considered as separator for compatibility with DOS texts.
 *  The search stops at a null character.
 */
char* FIND_WORD( ptr )
	char *ptr;		/* where the search should start */
{
	while (( *ptr ) && (( *ptr == ' ' ) || ( *ptr == '\t' ) || ( *ptr == '\r' ))) ptr ++;
	return( ptr );
}


/*********************************************************************************
 *  FIND_SEP()
 *
 *  FIND_SEP() searches for separator by skipping over non-separator characters.
 *  The carriage return is considered as separator for compatibility with DOS texts.
 *  The search stops at a null character.
 */
char* FIND_SEP( ptr )
	char *ptr;		/* where the search should start */
{
	while (( *ptr ) && (( *ptr != ' ' ) && ( *ptr != '\t' ) && ( *ptr != '\r' ))) ptr ++;
	return( ptr );
}


/*********************************************************************************
 *  COMPARE()
 *
 *  COMPARE() compares the word at ptr with the pattern pat.
 *  The carriage return is considered as separator for compatibility with DOS texts.
 *  Words longer than the pattern don't match.
 *
 *  Return values:
 *	1	... match
 *	0	... no match
 */
int COMPARE( ptr, pat )
	char *ptr;		/* pointer to the word to be compared */
	char *pat;		/* the pattern to compare with */
{
	int patlen;		/* length of the pattern string */
	char cterm;		/* character that should terminate the word */

	patlen = strlen( pat );
	cterm = *( ptr + patlen );
	if (( cterm ) && (( cterm != ' ' ) && ( cterm != '\t' ) && ( cterm != '\r' ))) return( 0 );
	return( strncmp( ptr, pat, patlen ) == 0 );
}


/*********************************************************************************
 *  NICE_NAME()
 *
 *  NICE_NAME() creates a new string derived from the passed string,
 *  by replacing underscore characters by spaces.
 *
 *  Return values:
 *	NULL	...	error occured
 *	! NULL	...	success, new string is returned
 */
char* NICE_NAME( name )
	char *name;		/* the original string */
{
	char *nice;		/* the new string */
	char *c;		/* pointer into nice */

	if (( nice = malloc( strlen( name ) + 1 )) == NULL ) {
		SYS_ERROR( "malloc() failed" );
		return( NULL );
	}
	strcpy( nice, name );

	c = nice;
	while ( *c ) {
		if ( *c == '_' ) *c = ' ';
		c ++;
	}
	return( nice );
}


/*********************************************************************************
 *  GET_WORD()
 *
 *  GET_WORD() reads a word, terminated by a separator, copies it
 *  into a newly allocated buffer and returns a pointer to that buffer.
 *
 *  Return values:
 *	NULL	...	error occured
 *	! NULL	...	success, new buffer is returned
 */
char* GET_WORD( ptr )
	char *ptr;		/* pointer to the word to be copied */
{
	char *buf;		/* the newly allocated buffer */
	char *sep;		/* pointer to the separator terminating the word */

	sep = FIND_SEP( ptr );
	if ( sep == ptr ) return( NULL );

	if (( buf = malloc( sep - ptr + 1 )) == NULL ) {
		SYS_ERROR( "malloc() failed" );
		return( NULL );
	}
	strncpy( buf, ptr, sep - ptr );
	*( buf + ( sep - ptr )) = 0;
	return( buf );
}


/*********************************************************************************
 *  REALLOC_OR_FREE()
 *
 *  REALLOC_OR_FREE() is just a frontend to realloc().
 *  When realloc() fails, it frees the original memory block.
 *
 *  Return value:
 *	NULL	...	failed or zero size requested
 *	! NULL	...	successful
 */
char* REALLOC_OR_FREE( orig, size )

	char *orig;		/* pointer to original memory block */
	long size;		/* the requested new size */
{
	char *new;		/* pointer to the new memory block */

	if (( new = realloc( orig, size )) == NULL ) {
		if ( orig ) free( orig );
		return( NULL );
	}
	return( new );
}


/*********************************************************************************
 *  GET_LINE()
 *
 *  GET_LINE() reads the next textual line from a file descriptor.
 *  The buffer must be either NULL or be allocated using malloc() before.
 *  The buffer is realloc'ed on demand in portions of 64 bytes.
 *  If the line is longer than 65536, error is returned.
 *  The buffer is freed if the end of file or error is encountered.
 *
 *  Return values:
 *	-2	...	error, the input line is too long
 *	-1	...	error occured
 *	 0	...	end of file or connection aborted
 *	> 0	...	characters read (incl. terminating \n, which is not copied)
 */
long GET_LINE( fd, buf )

	int fd;			/* file descriptor of the file to read from */
	char **buf;		/* pointer to pointer to the line buffer to read into */
{
	long i;			/* offset into the buffer */
	long size;		/* current size of the buffer */
	char c;			/* one character read from file */
	long result;		/* result of the read() call */

	i = 0;
	size = 0;

	while ( 1 ) {

		size += 64;
		if ( size > 65536 ) {
			APP_ERROR( "line too long" );
			free( *buf );
			return( -2 );
		}

		if (( *buf = REALLOC_OR_FREE( *buf, size )) == NULL ) {
			SYS_ERROR( "realloc() failed" );
			return( -1 );
		}

		while ( i < size ) {

			if (( result = read( fd, &c, 1 )) < 0 ) {
				SYS_ERROR( "read() failed" );
				if ( *buf ) free( *buf );
				return( -1 );
			} else if ( result == 0 ) {
				if ( *buf ) free( *buf );
				return( 0 );
			}

			if ( c == '\n' ) {
				*( *buf + i ) = 0;
				return( i + 1 );
			}
			*( *buf + i ) = c;
			i ++;
		}
	}
}


/*********************************************************************************
 *  INVERT_LIST()
 *
 *  INVERT_LIST() inverts the order of elements in the linked list and returns
 *  a pointer to the inverted list. The elements are not moved in memory.
 */
char* INVERT_LIST( last, offset )

	char *last;	/* pointer to the original linked list */
	int offset;	/* offset of the next field in the elements */
{
	char *first;	/* pointer to the inverted linked list */
	char *next;	/* pointer to the next element to be processed */

	first = NULL;
	while ( last ) {
		next = *((char **)( last + offset ));
		*((char **)( last + offset )) = first;
		first = last;
		last = next;
	}
	return( first );
}


/*********************************************************************************
 **	SECTION DEALING WITH KEYBOARD CONTROL PROTOCOL HANDLING			**
 ********************************************************************************/

/*********************************************************************************
 *  HANDLE_NOTIFY()
 *
 *  HANDLE_NOTIFY() handles one NOTIFY message arriving from Xks and reflects
 *  the information it carries in the Config structure. If the first line of
 *  the NOTIFY message was already read, it must be passed as first_line.
 *
 *  Return values:
 *	-1	...	error occured
 *	 0	...	completed successfully
 */
int HANDLE_NOTIFY( config, first_line )

	Config *config;		/* the Config structure */
	char *first_line;	/* first line of the message already read */
{
	char *ibuf;		/* buffer for the incoming reply */
	char *ptr;		/* pointer to a word in ibuf */
	KType *ktype;		/* the keyboard type to match to */


	DEBUG( "In HANDLE_NOTIFY" );

	/*
	 *  Get the first line of the message
	 */
	if ( first_line ) {
		if (( ibuf = malloc( strlen( first_line ) + 1 )) == NULL ) {
			SYS_ERROR( "malloc() failed" );
			return( -1 );
		}
		strcpy( ibuf, first_line );
	} else {
		ibuf = NULL;
		if ( GET_LINE( config -> ksock, &ibuf ) <= 0 ) return( -1 );
	}
	ptr = FIND_WORD( ibuf );

	/*
	 *  NOTIFY SWITCH TO message was received
	 */
	if ( COMPARE( ptr, "NOTIFY SWITCH TO" )) {

		DEBUG( "\tNOTIFY SWITCH TO message received" );
		ptr = FIND_WORD( FIND_SEP( ptr ));
		ptr = FIND_WORD( FIND_SEP( ptr ));
		ptr = FIND_WORD( FIND_SEP( ptr ));

		/*
		 *  If the list of keyboard types already exists,
		 *  go through it to find the matching KType
		 */
		if ( ktype = config -> first_ktype ) {
			while ( ktype ) {
				if ( COMPARE( ptr, ktype -> name )) break;
				ktype = (KType *) ktype -> next;
			}
			if ( ktype == NULL ) {
				APP_ERROR( "Keyboard type reported by Xks doesn't exist" );
				free( ibuf );
				return( -1 );
			}
			config -> active_ktype = ktype;
			DEBUG1( "\t( active_ktype == %s )", config -> active_ktype -> name );
		}

		/*
		 *  Get the second line of the message
		 */
		if ( GET_LINE( config -> ksock, &ibuf ) <= 0 ) return( -1 );

	/*
	 *  An unknown NOTIFY message was received
	 *  Ignore and read until an empty line is received
	 */
	} else if ( COMPARE( ptr, "NOTIFY" )) {
		DEBUG( "\tunknown NOTIFY message received" );

		while ( 1 ) {
			if ( GET_LINE( config -> ksock, &ibuf ) <= 0 ) return( -1 );
			ptr = FIND_WORD( ibuf );
			if ( ! *ptr ) break;
		}

	/*
	 *  Something else then a NOTIFY message was received
	 */
	} else {
		APP_ERROR1( "Unexpected message: %s", ibuf );
		free( ibuf );
		return( -1 );
	}

	free( ibuf );
	return( 0 );
}


/*********************************************************************************
 *  GET_REPLY_LINE()
 *
 *  GET_REPLY_LINE() retrieves one line from Xks which is a reply, that means
 *  which starts either with OK or ERR. Notify messages are filtered out
 *  in this function and are passed to HANDLE_NOTIFY().
 *  The buffer must be either NULL or be allocated using malloc() before.
 *  The buffer is freed if the end of file or error is encountered.
 *
 *  Return values:
 *	-2	...	error, the input line is too long
 *	-1	...	error occured
 *	 0	...	end of file or connection aborted
 *	> 0	...	characters read (incl. terminating \n, which is not copied)
 */
long GET_REPLY_LINE( config, ibuf )

	Config *config;		/* the Config structure */
	char **ibuf;		/* the line buffer */
{
	long result;		/* result of the GET_LINE() function */
	char *ptr;		/* pointer into *ibuf */

	while ( 1 ) {

		result = GET_LINE( config -> ksock, ibuf );
		if ( result <= 0 ) return( result );

		ptr = FIND_WORD( *ibuf );
		if ( COMPARE( ptr, "OK" ) || COMPARE( ptr, "ERR" )) break;

		if ( HANDLE_NOTIFY( config, *ibuf ) < 0 ) return( -1 );
	}
	return( result );
}


/*********************************************************************************
 *  LIST_REQUEST()
 *
 *  LIST_REQUEST() sends a LIST request to Xks, waits for the reply and
 *  then creates the list of keyboard types in the Config structure.
 *  LIST_REQUEST() should be run only once, as it assumes an empty list.
 *
 *  Return values:
 *	-1	...	error occured
 *	 0	...	completed successfully
 */
int LIST_REQUEST( config )

	Config *config;		/* the Config structure */
{
	char *obuf;		/* buffer for the outgoing request */
	long len;		/* length of obuf, without terminating NULL */
	char *ibuf;		/* buffer for the incoming reply */
	char *ptr;		/* pointer to a word in ibuf */
	char *name;		/* name of the keyboard type */
	KType *ktype;		/* the new keyboard type */


	DEBUG( "In LIST_REQUEST" );

	len = 6;
	if (( obuf = malloc( len + 1 )) == NULL ) {
		SYS_ERROR( "malloc() failed" );
		return( -1 );
	}
	if ( sprintf( obuf, "LIST\r\n" ) != len ) {
		APP_ERROR( "sprintf() != len" );
		free( obuf );
		return( -1 );
	}
	if ( write( config -> ksock, obuf, len ) != len ) {
		SYS_ERROR( "write() failed" );
		free( obuf );
		return( -1 );
	}
	free( obuf );

	/*
	 *  Wait for the first reply line
	 */
	ibuf = NULL;
	if ( GET_REPLY_LINE( config, &ibuf ) <= 0 ) return( -1 );
	ptr = FIND_WORD( ibuf );

	while ( 1 ) {

		if ( ! COMPARE( ptr, "OK LIST" )) {
			APP_ERROR( "Unexpected response from Xks" );
			free( ibuf );
			return( -1 );
		}
		ptr = FIND_WORD( FIND_SEP( ptr ));
		ptr = FIND_WORD( FIND_SEP( ptr ));

		if (( name = GET_WORD( ptr )) == NULL ) {
			APP_ERROR( "Unexpected response from Xks" );
			free( ibuf );
			return( -1 );
		}

		/*
		 *  Create a new KType and add it to the list in the Config structure
		 */
		if (( ktype = calloc( 1, sizeof( KType ))) == NULL ) {
			SYS_ERROR( "calloc() failed" );
			free( ibuf );
			free( name );
			return( -1 );
		}
		ktype -> next = (char *) config -> first_ktype;
		config -> first_ktype = ktype;

		ktype -> config = (char *) config;
		ktype -> name = name;
		if (( ktype -> nice_name = NICE_NAME( name )) == NULL ) {
			free( ibuf );
			return( -1 );
		}
		DEBUG2( "\tnew ktype: ( name == %s, nice_name == %s )", ktype -> name, ktype -> nice_name );

		/*
		 *  Get the next line
		 *  If we receive an empty line, the response is finished
		 */
		if ( GET_LINE( config -> ksock, &ibuf ) <= 0 ) return( -1 );
		ptr = FIND_WORD( ibuf );
		if ( ! *ptr ) break;
	}
	free( ibuf );

	/*
	 *  Invert the list of keyboard types
	 */
	config -> first_ktype = (KType *) INVERT_LIST( (char *) config -> first_ktype, OFFSET( KType*, next ));
	return( 0 );
}


/*********************************************************************************
 *  QUERY_REQUEST()
 *
 *  QUERY_REQUEST() sends a QUERY request to Xks, waits for the reply and
 *  then sets the active keyboard type in the Config structure.
 *
 *  Return values:
 *	-1	...	error occured
 *	 0	...	completed successfully
 */
int QUERY_REQUEST( config )

	Config *config;		/* the Config structure */
{
	char *obuf;		/* buffer for the outgoing request */
	long len;		/* length of obuf, without terminating NULL */
	char *ibuf;		/* buffer for the incoming reply */
	char *ptr;		/* pointer to a word in ibuf */
	KType *ktype;		/* the keyboard type to match to */


	DEBUG( "In QUERY_REQUEST" );

	len = 7;
	if (( obuf = malloc( len + 1 )) == NULL ) {
		SYS_ERROR( "malloc() failed" );
		return( -1 );
	}
	if ( sprintf( obuf, "QUERY\r\n" ) != len ) {
		APP_ERROR( "sprintf() != len" );
		free( obuf );
		return( -1 );
	}
	if ( write( config -> ksock, obuf, len ) != len ) {
		SYS_ERROR( "write() failed" );
		free( obuf );
		return( -1 );
	}
	free( obuf );

	/*
	 *  Wait for the first reply line
	 */
	ibuf = NULL;
	if ( GET_REPLY_LINE( config, &ibuf ) <= 0 ) return( -1 );
	ptr = FIND_WORD( ibuf );

	if ( ! COMPARE( ptr, "OK QUERY" )) {
		APP_ERROR( "Unexpected response from Xks" );
		free( ibuf );
		return( -1 );
	}
	ptr = FIND_WORD( FIND_SEP( ptr ));
	ptr = FIND_WORD( FIND_SEP( ptr ));

	/*
	 *  If the list of keyboard types already exists,
	 *  go through it to find the matching KType
	 */
	if ( ktype = config -> first_ktype ) {

		while ( ktype ) {
			if ( COMPARE( ptr, ktype -> name )) break;
			ktype = (KType *) ktype -> next;
		}
		if ( ktype == NULL ) {
			APP_ERROR( "Keyboard type reported by Xks doesn't exist" );
			free( ibuf );
			return( -1 );
		}
		config -> active_ktype = ktype;
		DEBUG1( "\t( active_ktype == %s )", config -> active_ktype -> name );
	}

	/*
	 *  Get the second reply line
	 */
	if ( GET_LINE( config -> ksock, &ibuf ) <= 0 ) return( -1 );
	free( ibuf );
	return( 0 );
}


/*********************************************************************************
 *  SWITCH_TO_REQUEST()
 *
 *  SWITCH_TO_REQUEST() sends a SWITCH TO request to Xks and waits for the reply.
 *
 *  Return values:
 *	-1	...	error occured
 *	 0	...	completed successfully
 */
int SWITCH_TO_REQUEST( config, ktype )

	Config *config;		/* the Config structure */
	KType *ktype;		/* keyboard type to switch to */
{
	char *obuf;		/* buffer for the outgoing request */
	long len;		/* length of obuf, without terminating NULL */
	char *ibuf;		/* buffer for the incoming reply */
	char *ptr;		/* pointer to a word in ibuf */


	DEBUG( "In SWITCH_TO_REQUEST" );

	len = 12 + strlen( ktype -> name );
	if (( obuf = malloc( len + 1 )) == NULL ) {
		SYS_ERROR( "malloc() failed" );
		return( -1 );
	}
	if ( sprintf( obuf, "SWITCH TO %s\r\n", ktype -> name ) != len ) {
		APP_ERROR( "sprintf() != len" );
		free( obuf );
		return( -1 );
	}
	if ( write( config -> ksock, obuf, len ) != len ) {
		SYS_ERROR( "write() failed" );
		free( obuf );
		return( -1 );
	}
	free( obuf );

	/*
	 *  Wait for the first reply line
	 */
	ibuf = NULL;
	if ( GET_REPLY_LINE( config, &ibuf ) <= 0 ) return( -1 );
	ptr = FIND_WORD( ibuf );

	if ( ! COMPARE( ptr, "OK SWITCH TO" )) {
		APP_ERROR( "Unexpected response from Xks" );
		free( ibuf );
		return( -1 );
	}

	/*
	 *  Get the second reply line
	 */
	if ( GET_LINE( config -> ksock, &ibuf ) <= 0 ) return( -1 );
	free( ibuf );

	/*
	 *  Set the new active keyboard type
	 */
	config -> active_ktype = ktype;
	DEBUG1( "\t( active_ktype == %s )", config -> active_ktype -> name );
	return( 0 );
}


/*********************************************************************************
 **		SECTION RELATED TO THE MAIN FUNCTION				**
 ********************************************************************************/

/*********************************************************************************
 *  SELF_DIR()
 *
 *  This function is used to find the directory where the executable
 *  of a running program resides.
 *
 *  Return value:
 *	string	...	directory, caller responsible for freeing
 *	NULL	...	error, insufficient memory
 */
char* SELF_DIR( argv0 )

	char *argv0;			/* argv[0] of the program whose directory we are looking for */
{
	char *dir;			/* pointer to the directory that will be returned */
	char *sep;			/* pointer into dir to the slash separating directory from file */
	char *path;			/* pointer to the start of a PATH component */
	char *p;			/* pointer to the end of a PATH component */
	char *dpath = DEFAULT_PATH;	/* default path if PATH unset */
	uid_t euid;			/* effective used id */
	gid_t egid;			/* effective group id */
	struct stat st;			/* structure returned by stat() */
	unsigned int bit;		/* permissions mask */
	char buff[ LINK_LEN ];		/* string buffer */
	long len;			/* length variable */


	/*
	 *  Copy argv0 to dir, so that we have got a valid intermediate
	 *  result in case argv0 contains one or more slashes
	 */
	len = strlen( argv0 );

	if (( dir = malloc( len + 1 )) == NULL ) {
		SYS_ERROR( "malloc() failed" );
		return( NULL );
	}

	strcpy( dir, argv0 );
	if (( sep = strrchr( dir, '/' )) == NULL ) {

		/*
		 *  argv0 contains no slashes
		 *  Use the PATH environment variable to find file executable by us.
		 *  If PATH is unset, use default path.
		 */
		if (( path = getenv( "PATH" )) == NULL ) path = dpath;

		if (( dir = REALLOC_OR_FREE( dir, strlen( path ) + len + 2 )) == NULL ) {
			SYS_ERROR( "realloc() failed" );
			return( NULL );
		}
		euid = geteuid();
		egid = getegid();

		p = path;
		do {
			path = p;
			if (( p = strchr( path, ':' )) == NULL ) {
				p = strchr( path, 0 );
			}

			/*
			 *  Construct the absolute filename we want to test
			 *  Two adjacent colons, or a colon at the beginning or the end
			 *  of PATH means to search the current directory
			 */
			if ( p == path ) {
				strcpy( dir, argv0 );
				sep = NULL;
			} else {
				strncpy( dir, path, p - path );
				sep = dir + ( p - path );
				*sep = '/';
				strcpy( sep + 1, argv0 );
			}

			/*
			 *  Test if the file exists and is executable by us
			 *  If the file is a symbolic link, stat() tests the file pointed
			 *  to by this link.
			 */
			if (( stat( dir, &st ) == 0 ) && S_ISREG( st.st_mode )) {

				if ( st.st_uid == euid ) bit = S_IXUSR;
				else if ( st.st_gid == egid ) bit = S_IXGRP;
				else bit = S_IXOTH;

				if ( st.st_mode & bit ) goto self_dir_found;
			}

		} while ( *p ++ );

		/*
		 *  We haven't found argv0 in path.
		 *  The maximum what we can do is to assume argv0 in current directory
		 */
		strcpy( dir, argv0 );
		sep = NULL;
	}

self_dir_found:

	/*
	 *  The file we have found up to now can be a symbolic link.
	 *  We have to find the file it points to, possibly through a
	 *  cascade of other links.
	 */
	while (( len = readlink( dir, buff, sizeof( buff ))) > 0 ) {

		/*
		 *  If the link points to an absolute pathname,
		 *  replace dir by the result of readlink().
		 *  The same applies if link doesn't contain a
		 *  separating slash (ie. link lies in current directory)
		 */
		if (( buff[ 0 ] == '/' ) || ( sep == NULL )) {

			if (( dir = REALLOC_OR_FREE( dir, len + 1 )) == NULL ) {
				SYS_ERROR( "realloc() failed" );
				return( NULL );
			}
			strncpy( dir, buff, len );
			*( dir + len ) = 0;

		/*
		 *  The link points to a relative pathname and link
		 *  contains a separating slash.
		 *  Replace the file portion of link by the result of readlink().
		 */
		} else {

			if (( dir = REALLOC_OR_FREE( dir, ( sep - dir ) + len + 2 )) == NULL ) {
				SYS_ERROR( "realloc() failed" );
				return( NULL );
			}
			sep = strrchr( dir, '/' );
			strncpy( sep + 1, buff, len );
			*( sep + len + 1 ) = 0;
		}
		sep = strrchr( dir, '/' );
	}

	/*
	 *  Return the directory of the file in the dir string.
	 *  If the separating slash is the first character of dir, the directory is /
	 *  If the separating slash is somewhere else, terminate dir there
	 *  If there is no separating slash, return the current directory
	 */
	if ( sep == dir ) {
		strcpy( dir, "/" );
	} else if ( sep ) {
		*sep = 0;
	} else {
		strcpy( dir, "." );
	}

	return( dir );
}


/*********************************************************************************
 *  PARSE_DISPLAY()
 *
 *  The PARSE_DISPLAY() function is used to parse string variables with
 *  DISPLAY-like syntax. It extracts the IP address, port number and screen number.
 *  If the variable doesn't contain the IP address, then our_ipaddr is used.
 *
 *  Return value:
 *	 0	...	the parsing went OK
 *	-1	...	error occured, message was printed
 */
int PARSE_DISPLAY( variable, ipaddr, port, screen, our_ipaddr )

	char *variable;			/* the variable's value, returned by getenv() */
	struct in_addr *ipaddr;		/* result: the IP address, MSB byte order */
	unsigned short *port;		/* result: the port number, host byte order */
	int *screen;			/* result: the screen number, host byte order */
	struct in_addr our_ipaddr;	/* our IP address, MSB byte order */
{
	char *var;			/* our copy of the variables's value */
	char *colon;			/* position of the colon in var */
	char *dot;			/* position of the dot in var */
	struct hostent *host_ent;	/* hostent of the machine with ipaddr */

	if (( var = malloc( strlen( variable ) + 1 )) == NULL ) {
		SYS_ERROR( "malloc() failed" );
		return( -1 );
	}
	strcpy( var, variable );

	if (( colon = strchr( var, ':' )) == NULL ) {
		APP_ERROR( "Variable doesn't contain a colon" );
		goto parse_display_fail;
	}
	*colon = 0;

	dot = strchr( colon + 1, '.' );
	if ( dot ) *dot = 0;

	if ( colon == var ) {
		*ipaddr = our_ipaddr;
	} else {
		if (( host_ent = gethostbyname( var )) == NULL ) {
			SYS_ERROR( "gethostbyname() failed" );
			goto parse_display_fail;
		}
		*ipaddr = *( (struct in_addr *) host_ent -> h_addr );
	}

	if ( sscanf( colon + 1, "%hd", port ) != 1 ) {
		APP_ERROR( "Unable to parse port number" );
		goto parse_display_fail;
	}
	*port += 6000;

	if ( dot ) {
		if ( sscanf( dot + 1, "%d", screen ) != 1 ) {
			APP_ERROR( "Unable to parse screen number" );
			goto parse_display_fail;
		}
	} else {
		*screen = 0;
	}

	free( var );
	return( 0 );

parse_display_fail:
	free( var );
	return( -1 );
}


/*********************************************************************************
 *  TOGGLE_CALLBACK()
 *
 *  This function is called by X Toolkit when the state of a ToggleButton changes.
 */
void TOGGLE_CALLBACK( widget, client_data, call_data )

	Widget widget;			/* unused */
	XtPointer client_data;		/* Client data, pointer to the KType structure */
	XtPointer call_data;		/* Call data, info about the state of the ToggleButton */
{
	Config *config;			/* the Config structure */
	KType *ktype;			/* the KType to which to set the active keyboard type */

#ifndef USE_XAW
	XmToggleButtonCallbackStruct *state;	/* info about the state of the ToggleButton */
#endif

	DEBUGN( "In TOGGLE_CALLBACK: " );
	ktype = (KType *) client_data;
	config = (Config *) ktype -> config;


	/*
	 *  We are called two times: once due the old ToggleButton being de-activated
	 *  and once due the new ToggleButton being activated. The first "if" filters
	 *  out the second situation.
	 *
	 *  Further, we can be called both when the state of the ToggleButton was changed
	 *  by the user and when it was changed from this program. The second "if" filters
	 *  out the first situation.
	 *
	 *  Then, if we passed both tests, ie. the new ToggleButton was activated by the user,
	 *  call SWITCH_TO_REQUEST() to tell Xks to switch to the new keyboard type.
	 */
#ifndef USE_XAW

	state = (XmToggleButtonCallbackStruct *) call_data;
	if ( state -> set ) {
		DEBUGN( "ToggleButton activated: " );
		if ( ktype != config -> active_ktype ) {
			DEBUG( "by the user" );
			if ( SWITCH_TO_REQUEST( config, ktype ) < 0 ) exit( -1 );
		}
	}
#else
	if ( ktype == XawToggleGetCurrent( config -> first_ktype -> toggle )) {
		DEBUGN( "ToggleButton activated: " );
		if ( ktype != config -> active_ktype ) {
			DEBUG( "by the user" );
			if ( SWITCH_TO_REQUEST( config, ktype ) < 0 ) exit( -1 );
		}
	}
#endif
	DEBUGNL();
}


/*********************************************************************************
 *  NOTIFY_CALLBACK()
 *
 *  This function is called by X Toolkit whenever there is input pending
 *  on the socket to Xks, ie. when a NOTIFY message arrives
 */
void NOTIFY_CALLBACK( client_data, fd, id )

	XtPointer client_data;		/* Client data, pointer to the Config structure */
	int *fd;			/* unused */
	XtInputId *id;			/* unused */
{
	Config *config;			/* the Config structure */
	KType *old_ktype;		/* the old keyboard type */

	DEBUG( "In NOTIFY_CALLBACK" );
	config = (Config *) client_data;

	/*
	 *  Remember the old keyboard type in old_ktype
	 *  Call HANDLE_NOTIFY to retrieve the new active keyboard type from the NOTIFY message
	 */
	old_ktype = config -> active_ktype;
	if ( HANDLE_NOTIFY( config, NULL ) < 0 ) exit( -1 );
	DEBUGN( "In NOTIFY_CALLBACK: " );

	/*
	 *  If the keyboard type really changed, change the state of the ToggleButtons
	 */
	if ( config -> active_ktype != old_ktype ) {
		DEBUG( "active keyboard type really changed" );

#ifndef USE_XAW
		XmToggleButtonGadgetSetState( old_ktype -> toggle, 0, 0 );
		XmToggleButtonGadgetSetState( config -> active_ktype -> toggle, 1, 0 );
#else
		XawToggleSetCurrent( config -> first_ktype -> toggle, config -> active_ktype );
#endif
	}
	DEBUGNL();
}


/*********************************************************************************
 *  INTERACTIVE()
 *
 *  This function builds and runs the interactive Motif GUI.
 *  It never returns except in case of error.
 *
 *  Return values:
 *	-1	...	error occured
 */
int INTERACTIVE( config, app, toplevel )

	Config *config;		/* the Config structure */
	XtAppContext app;	/* X Toolkit application context */
	Widget toplevel;	/* X Toolkit application top level widget */
{
	Widget radio_box;	/* the RadioBox widget */
	KType *ktype;		/* keyboard type for which ToggleButton is created */
	Arg args[ 3 ];		/* resource-value array */
	int args_n;		/* number of elements in args */

#ifndef USE_XAW
	XmString str;		/* Motif compound string */
#endif

	/*
	 *  Send a LIST request to Xks to get the list of keyboard types
	 */
	if ( LIST_REQUEST( config ) < 0 ) return( -1 );

	/*
	 *  Build the user interface
	 *  Go through the keyboard types and create a ToggleButton for each
	 */
#ifndef USE_XAW
	radio_box = XmCreateRadioBox( toplevel, RADIO_BOX, NULL, 0 );
#else
	radio_box = XtCreateManagedWidget( RADIO_BOX, boxWidgetClass, toplevel, NULL, 0 );
#endif

	ktype = config -> first_ktype;
	while ( ktype ) {

#ifndef USE_XAW
		args_n = 0;
		str = XmStringCreateLocalized( ktype -> nice_name );
		XtSetArg( args[ args_n ], XmNlabelString, (XtArgVal) str ); args_n ++;
		ktype -> toggle = XtCreateManagedWidget( ktype -> name,
			xmToggleButtonGadgetClass, radio_box, args, args_n );
		XmStringFree( str );

		XtAddCallback( ktype -> toggle, XmNvalueChangedCallback,
			TOGGLE_CALLBACK, (XtPointer) ktype );
#else
		args_n = 0;
		XtSetArg( args[ args_n ], XtNlabel, ktype -> nice_name ); args_n ++;
		XtSetArg( args[ args_n ], XtNradioData, ktype ); args_n ++;
		if ( ktype != config -> first_ktype ) {
			XtSetArg( args[ args_n ], XtNradioGroup, config -> first_ktype -> toggle ); args_n ++;
		}
		ktype -> toggle = XtCreateManagedWidget( ktype -> name,
			toggleWidgetClass, radio_box, args, args_n );

		XtAddCallback( ktype -> toggle, XtNcallback,
			TOGGLE_CALLBACK, (XtPointer) ktype );
#endif

		ktype = (KType *) ktype -> next;
	}

	/*
	 *  Send a QUERY request to Xks to get the active keyboard type
	 */
	if ( QUERY_REQUEST( config ) < 0 ) return( -1 );

	/*
	 *  Set the state of the ToggleButton that represents the active keyboard type.
	 */
#ifndef USE_XAW
	XmToggleButtonGadgetSetState( config -> active_ktype -> toggle, 1, 0 );
#else
	XawToggleSetCurrent( config -> first_ktype -> toggle, config -> active_ktype );
#endif

	/*
	 *  Add the socket input to the X Toolkit, to be able to react to NOTIFY messages
	 */
	XtAppAddInput( app, config -> ksock, (XtPointer) XtInputReadMask,
		NOTIFY_CALLBACK, (XtPointer) config );

	/*
	 *  Run the user interface
	 */
	XtManageChild( radio_box );
	XtRealizeWidget( toplevel );
	XtAppMainLoop( app );
}


/*********************************************************************************
 *  QUERY()
 *
 *  This function queries Xks for the active keyboard type and prints
 *  its name to standard output.
 *
 *  Return values:
 *	-1	...	error occured
 *	 0	...	successful
 */
int QUERY( config )
	Config *config;		/* the Config structure */
{
	/*
	 *  Send a LIST request to Xks to get the list of keyboard types
	 *  Send a QUERY request to Xks to get the active keyboard type
	 */
	if ( LIST_REQUEST( config ) < 0 ) return( -1 );
	if ( QUERY_REQUEST( config ) < 0 ) return( -1 );

	fprintf( stdout, "%s\n", config -> active_ktype -> name );
	return( 0 );
}


/*********************************************************************************
 *  SWITCHTO()
 *
 *  This function tells Xks to switch to the specified keyboard type.
 *
 *  Return values:
 *	-1	...	error occured
 *	 0	...	successful
 */
int SWITCHTO( config, argv_switchto )

	Config *config;		/* the Config structure */
	char *argv_switchto;	/* name of the keyboard type to switch to */
{
	KType *ktype;		/* keyboard type to switch to */

	/*
	 *  Send a LIST request to Xks to get the list of keyboard types
	 */
	if ( LIST_REQUEST( config ) < 0 ) return( -1 );

	/*
	 *  Lookup the specified keyboard type
	 */
	ktype = config -> first_ktype;
	while ( ktype ) {
		if ( COMPARE( argv_switchto, ktype -> name )) break;
		ktype = (KType *) ktype -> next;
	}
	if ( ktype == NULL ) {
		APP_ERROR( "The specified keyboard type doesn't exist" );
		return( -1 );
	}

	/*
	 *  Send a SWITCH TO request to set the specified keyboard type
	 */
	if ( SWITCH_TO_REQUEST( config, ktype ) < 0 ) return( -1 );
	return( 0 );
}


/*********************************************************************************
 *  main()
 *
 *  Finally, the main() function.
 *  We can be called with the following command-line arguments:
 *
 *	xksclient [ -query | -switchto <ktype> ] [ <X Toolkit options> ]
 *
 *  With the -query option, we ask Xks for the name of the active keyboard
 *  type and print it to stdout.
 *
 *  With the -switchto <ktype> option, we tell Xks to switch to the specified
 *  keyboard type.
 *
 *  If neither option is specified, we run an interactive Motif GUI.
 *
 *  Exit code:
 *	-1	...	error occured
 *	 0	...	completed successfully
 */
int main( argc, argv )

	int argc;				/* number of command line arguments */
	char *argv[];				/* array of command line arguments */
{
	int argv_query;				/* if == 1, the -query command line option is specified */
	char *argv_switchto;			/* if != NULL, value of -switchto command line option */
	char **ptr1;				/* pointer to an element of the argv list */
	char **ptr2;				/* pointer to an element of the argv list */

	Config *config;				/* the Config structure */

	char *selfdir;				/* directory where our executable resides */
	char xapplresdir[ LINK_LEN ];		/* new value of the XAPPLRESDIR environment variable */

	XtAppContext app;			/* X Toolkit application context */
	Widget toplevel;			/* X Toolkit application top level widget */
	Display *display;			/* X Server connection via Xlib */

	Atom keyboard_atom;			/* atom of the KEYBOARD PROPERTY property */
	char *keyboard_val;			/* value of the KEYBOARD_PROPERTY property */
	Atom actual_type;			/* return value of GetWindowProperties() */
	int actual_format;			/* return value of GetWindowProperties() */
	unsigned long nitems;			/* return value of GetWindowProperties() */
	unsigned long bytes_after;		/* return value of GetWindowProperties() */

	struct in_addr k_ipaddr;		/* IP address of Xks, MSB byte order */
	unsigned short k_port;			/* port number of Xks, host byte order */
	int screen_dummy;			/* screen number, dummy variable */
	struct in_addr ip_dummy;		/* IP address, dummy variable */

	struct sockaddr_in addr;		/* socket address structure for connect() */

	char *ibuf;				/* buffer for the WELCOME message */
	int result;				/* result of the GET_LINE() function */


	/*
	 *  We go through the array of command line arguments and search for
	 *  one of the following two options, then remove it from argv.
	 *  If both options are specified, we report an error and exit.
	 *
	 *	-query
	 *	-switchto <ktype>
	 */
	argv_query = 0;
	argv_switchto = NULL;

	ptr1 = argv;
	ptr2 = argv;

	while ( *ptr2 = *ptr1 ) {

		if ( COMPARE( *ptr1, "-query" )) {
			argv_query = 1;
			ptr1 ++;
			argc -= 1;
			DEBUG( "( argv_query == 1 )" );

		} else if ( COMPARE( *ptr1, "-switchto" )) {
			ptr1 ++;
			if ( *ptr1 == NULL ) goto main_argv_usage;
			argv_switchto = *ptr1;
			ptr1 ++;
			argc -= 2;
			DEBUG1( "( argv_switchto == %s )", argv_switchto );

		} else {
			ptr1 ++;
			ptr2 ++;
		}
	}
	if ( argv_query && argv_switchto ) goto main_argv_usage;


	/*
	 *  Create the Config structure
	 */
	if (( config = calloc( 1, sizeof( Config ))) == NULL ) {
		SYS_ERROR( "calloc() failed" );
		exit( -1 );
	}


	/*
	 *  Figure out the directory where our executable resides and set
	 *  XAPPLRESDIR to point to this directory, so that the X Resource file
	 *  that is in this directory is used.
	 */
	if (( selfdir = SELF_DIR( argv[0] )) == NULL ) exit( -1 );
	DEBUG1( "( selfdir == %s )", selfdir );

	if ( sprintf( xapplresdir, "XAPPLRESDIR=%s", selfdir ) >= sizeof( xapplresdir )) {
		APP_ERROR( "not enough space for sprintf()" );
		exit( -1 );
	}
	if ( putenv( xapplresdir ) < 0 ) {
		SYS_ERROR( "putenv() failed" );
		exit( -1 );
	}


	/*
	 *  Initialize the X Toolkit and connect to the X server
	 *  Obtain the Display pointer, as needed for the Xlib calls
	 */
	XtSetLanguageProc( NULL, NULL, NULL );
	toplevel = XtAppInitialize( &app, APPLICATION_CLASS, NULL, 0, &argc, argv, NULL, NULL, 0 );
	display = XtDisplay( toplevel );


	/*
	 *  Read the value of the KEYBOARD_PROPERTY property
	 *  If the property is not set, Xks is not running, we report an error message and exit
	 */
	if (( keyboard_atom = XInternAtom( display, KEYBOARD_PROPERTY, False )) == None ) {
		APP_ERROR( "XInternAtom() failed" );
		exit( -1 );
	}
	if ( XGetWindowProperty( display, RootWindow( display, 0 ),
		keyboard_atom, 0, DISP_LEN / 4, False, XA_STRING,
		&actual_type, &actual_format, &nitems, &bytes_after,
		(unsigned char**) &keyboard_val ) != Success ) {

		APP_ERROR( "XGetWindowProperty() failed" );
		exit( -1 );
	}
	if ( nitems == 0 ) keyboard_val = NULL;
	DEBUG1( "( keyboard_val == %s )", keyboard_val ? keyboard_val : "NULL" );

	if ( keyboard_val == NULL ) {
		APP_ERROR( "Xks is not running on this display" );
		exit( -1 );
	}


	/*
	 *  Parse the value of the KEYBOARD_PROPERTY property, to obtain information
	 *  about how to connect to Xks.
	 */
	if ( PARSE_DISPLAY( keyboard_val, &k_ipaddr, &k_port, &screen_dummy, ip_dummy ) < 0 ) {
		APP_ERROR1( "Unable to parse keyboard_val: %s", keyboard_val );
		exit( -1 );
	}
	DEBUG2( "( k_ipaddr == %s, k_port == %hd )", inet_ntoa( k_ipaddr ), k_port );


	/*
	 *  Create a socket and connect it to the keyboard control port of Xks
	 */
	if (( config -> ksock = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
		SYS_ERROR( "socket() failed" );
		exit( -1 );
	}
	memset( &addr, 0, sizeof( addr ));
	addr.sin_family = AF_INET;
	addr.sin_addr = k_ipaddr;
	addr.sin_port = htons( k_port );

	if ( connect( config -> ksock, (struct sockaddr*) &addr, sizeof( addr )) < 0 ) {
		SYS_ERROR( "connect() failed" );
		APP_ERROR( "Xks is not running on this display" );
		exit( -1 );
	}


	/*
	 *  Retrieve the WELCOME message from Xks
	 *
	 *  Remember that if we connect to wrong software with wrong protocol here,
	 *  we might potentially hang forever in GET_LINE().
	 */
	ibuf = NULL;
	if (( result = GET_LINE( config -> ksock, &ibuf )) == -1 ) exit( -1 );
	if (( result == 0 ) || ( result == -2 )) {
		APP_ERROR1( "Other software than Xks on %s", keyboard_val );
		exit( -1 );
	}

	DEBUG1( "( ibuf == %s )", ibuf );
	if ( ! COMPARE( ibuf, "WELCOME" )) {
		APP_ERROR1( "Other software than Xks on %s", keyboard_val );
		exit( -1 );
	}

	/*
	 *  Get the empty line that terminates the WELCOME message
	 */
	if ( GET_LINE( config -> ksock, &ibuf ) <= 0 ) exit( -1 );
	free( ibuf );


	/*
	 *  See if we have to be interactive or not and what we have to do
	 */
	if ( argv_query ) {
		DEBUG( "Entering QUERY" );
		exit( QUERY( config ));
	} else if ( argv_switchto ) {
		DEBUG( "Entering SWITCHTO" );
		exit( SWITCHTO( config, argv_switchto ));
	} else {
		DEBUG( "Entering INTERACTIVE" );
		exit( INTERACTIVE( config, app, toplevel ));
	}


main_argv_usage:

	fprintf( stderr, "Usage:\n" );
	fprintf( stderr, "    xksclient [ -query | -switchto <ktype> ] [ <X Toolkit options> ]\n" );
	exit( -1 );
}

/* End */
